import warnings
from coffea.nanoevents import transforms
from coffea.nanoevents.util import quote, concat
from .base import BaseSchema, listarray_form, zip_forms, nest_jagged_forms


class NanoAODSchema(BaseSchema):
    """NanoAOD schema builder

    The NanoAOD schema is built from all branches found in the supplied file, based on
    the naming pattern of the branches. The following additional arrays are constructed:

    - Any branches named ``n{name}`` are assumed to be counts branches and converted to offsets ``o{name}``
    - Any local index branches with names matching ``{source}_{target}Idx*`` are converted to global indexes for the event chunk (postfix ``G``)
    - Any `nested_items` are constructed, if the necessary branches are available
    - Any `special_items` are constructed, if the necessary branches are available

    From those arrays, NanoAOD collections are formed as collections of branches grouped by name, where:

    - one branch exists named ``name`` and no branches start with ``name_``, interpreted as a single flat array;
    - one branch exists named ``name``, one named ``n{name}``, and no branches start with ``name_``, interpreted as a single jagged array;
    - no branch exists named ``{name}`` and many branches start with ``name_*``, interpreted as a flat table; or
    - one branch exists named ``n{name}`` and many branches start with ``name_*``, interpreted as a jagged table.

    Collections are assigned mixin types according to the `mixins` mapping.
    All collections are then zipped into one `base.NanoEvents` record and returned.

    There is a class-level variable ``warn_missing_crossrefs`` which will alter the behavior of
    NanoAODSchema. If warn_missing_crossrefs is true then when a missing global index cross-ref
    target is encountered a warning will be issued instead of an exception. (Default: False)
    """

    warn_missing_crossrefs = False

    mixins = {
        "CaloMET": "MissingET",
        "ChsMET": "MissingET",
        "GenMET": "MissingET",
        "MET": "MissingET",
        "METFixEE2017": "MissingET",
        "PuppiMET": "MissingET",
        "RawMET": "MissingET",
        "TkMET": "MissingET",
        # pseudo-lorentz: pt, eta, phi, mass=0
        "IsoTrack": "PtEtaPhiMCollection",
        "SoftActivityJet": "PtEtaPhiMCollection",
        "TrigObj": "PtEtaPhiMCollection",
        # True lorentz: pt, eta, phi, mass
        "FatJet": "FatJet",
        "GenDressedLepton": "PtEtaPhiMCollection",
        "GenJet": "PtEtaPhiMCollection",
        "GenJetAK8": "FatJet",
        "Jet": "Jet",
        "LHEPart": "PtEtaPhiMCollection",
        "SV": "PtEtaPhiMCollection",
        "SubGenJetAK8": "PtEtaPhiMCollection",
        "SubJet": "PtEtaPhiMCollection",
        # Candidate: lorentz + charge
        "Electron": "Electron",
        "Muon": "Muon",
        "Photon": "Photon",
        "FsrPhoton": "FsrPhoton",
        "Tau": "Tau",
        "GenVisTau": "GenVisTau",
        # special
        "GenPart": "GenParticle",
    }
    """Default configuration for mixin types, based on the collection name.

    The types are implemented in the `coffea.nanoevents.methods.nanoaod` module.
    """
    nested_items = {
        "FatJet_subJetIdxG": ["FatJet_subJetIdx1G", "FatJet_subJetIdx2G"],
        "Jet_muonIdxG": ["Jet_muonIdx1G", "Jet_muonIdx2G"],
        "Jet_electronIdxG": ["Jet_electronIdx1G", "Jet_electronIdx2G"],
    }
    """Default nested collections, where nesting is accomplished by a fixed-length set of indexers"""
    special_items = {
        "GenPart_distinctParentIdxG": (
            transforms.distinctParent_form,
            ("GenPart_genPartIdxMotherG", "GenPart_pdgId"),
        ),
        "GenPart_childrenIdxG": (
            transforms.children_form,
            (
                "oGenPart",
                "GenPart_genPartIdxMotherG",
            ),
        ),
        "GenPart_distinctChildrenIdxG": (
            transforms.children_form,
            (
                "oGenPart",
                "GenPart_distinctParentIdxG",
            ),
        ),
    }
    """Default special arrays, where the callable and input arrays are specified in the value"""

    def __init__(self, base_form, version="6"):
        self._version = version
        super().__init__(base_form)
        self._form["contents"] = self._build_collections(self._form["contents"])
        self._form["parameters"]["metadata"]["version"] = self._version

    def _build_collections(self, branch_forms):
        # parse into high-level records (collections, list collections, and singletons)
        collections = set(k.split("_")[0] for k in branch_forms)
        collections -= set(
            k for k in collections if k.startswith("n") and k[1:] in collections
        )

        # Create offsets virtual arrays
        for name in collections:
            if "n" + name in branch_forms:
                branch_forms["o" + name] = transforms.counts2offsets_form(
                    branch_forms["n" + name]
                )

        # Create global index virtual arrays for indirection
        idxbranches = [k for k in branch_forms if "Idx" in k]
        for name in collections:
            indexers = [k for k in idxbranches if k.startswith(name + "_")]
            for k in indexers:
                target = k[len(name) + 1 : k.find("Idx")]
                target = target[0].upper() + target[1:]
                if target not in collections:
                    problem = RuntimeError(
                        "Parsing indexer %s, expected to find collection %s but did not"
                        % (k, target)
                    )
                    if self.__class__.warn_missing_crossrefs:
                        warnings.warn(str(problem), RuntimeWarning)
                    else:
                        raise problem
                branch_forms[k + "G"] = transforms.local2global_form(
                    branch_forms[k], branch_forms["o" + target]
                )

        # Create nested indexer from Idx1, Idx2, ... arrays
        for name, indexers in self.nested_items.items():
            if all(idx in branch_forms for idx in indexers):
                branch_forms[name] = transforms.nestedindex_form(
                    [branch_forms[idx] for idx in indexers]
                )

        # Create any special arrays
        for name, (fcn, args) in self.special_items.items():
            if all(k in branch_forms for k in args):
                branch_forms[name] = fcn(*(branch_forms[k] for k in args))

        output = {}
        for name in collections:
            mixin = self.mixins.get(name, "NanoCollection")
            if "o" + name in branch_forms and name not in branch_forms:
                # list collection
                offsets = branch_forms["o" + name]
                content = {
                    k[len(name) + 1 :]: branch_forms[k]
                    for k in branch_forms
                    if k.startswith(name + "_")
                }
                output[name] = zip_forms(
                    content, name, record_name=mixin, offsets=offsets
                )
                output[name]["content"]["parameters"].update(
                    {
                        "__doc__": offsets["parameters"]["__doc__"],
                        "collection_name": name,
                    }
                )
            elif "o" + name in branch_forms:
                # list singleton, can use branch's own offsets
                output[name] = branch_forms[name]
                output[name]["parameters"].update(
                    {"__array__": mixin, "collection_name": name}
                )
            elif name in branch_forms:
                # singleton
                output[name] = branch_forms[name]
            else:
                # simple collection
                output[name] = zip_forms(
                    {
                        k[len(name) + 1 :]: branch_forms[k]
                        for k in branch_forms
                        if k.startswith(name + "_")
                    },
                    name,
                    record_name=mixin,
                )
                output[name]["parameters"].update({"collection_name": name})

        return output

    @property
    def behavior(self):
        """Behaviors necessary to implement this schema"""
        from coffea.nanoevents.methods import nanoaod

        return nanoaod.behavior



import warnings
from cachetools import LRUCache
from collections.abc import Mapping
import uproot
import awkward
import numpy
import json
from coffea.nanoevents.mapping.base import UUIDOpener, BaseSourceMapping
from coffea.nanoevents.util import quote, key_to_tuple, tuple_to_key


class TrivialUprootOpener(UUIDOpener):
    def __init__(self, uuid_pfnmap, uproot_options={}):
        super(TrivialUprootOpener, self).__init__(uuid_pfnmap)
        self._uproot_options = uproot_options

    def open_uuid(self, uuid):
        pfn = self._uuid_pfnmap[uuid]
        rootdir = uproot.open(pfn, **self._uproot_options)
        if str(rootdir.file.uuid) != uuid:
            raise RuntimeError(
                f"UUID of file {pfn} does not match expected value ({uuid})"
            )
        return rootdir


class UprootSourceMapping(BaseSourceMapping):
    _debug = False

    def __init__(self, fileopener, cache=None, access_log=None):
        super(UprootSourceMapping, self).__init__(fileopener, cache, access_log)

    @classmethod
    def _extract_base_form(cls, tree):
        branch_forms = {}
        for key, branch in tree.iteritems():
            if "," in key or "!" in key:
                warnings.warn(
                    f"Skipping {key} because it contains characters that NanoEvents cannot accept [,!]"
                )
                continue
            if len(branch):
                continue
            form = branch.interpretation.awkward_form(None)
            form = uproot._util.awkward_form_remove_uproot(awkward, form)
            form = json.loads(form.tojson())
            if (
                form["class"].startswith("ListOffset")
                and form["content"]["class"] == "NumpyArray"  # noqa
            ):
                form["form_key"] = quote(f"{key},!load")
                form["content"]["form_key"] = quote(f"{key},!load,!content")
                form["content"]["parameters"] = {"__doc__": branch.title}
            elif form["class"] == "NumpyArray":
                form["form_key"] = quote(f"{key},!load")
                form["parameters"] = {"__doc__": branch.title}
            else:
                warnings.warn(
                    f"Skipping {key} as it is not interpretable by NanoEvents"
                )
                continue
            branch_forms[key] = form

        return {
            "class": "RecordArray",
            "contents": branch_forms,
            "parameters": {"__doc__": tree.title},
            "form_key": "",
        }

    def key_root(self):
        return "UprootSourceMapping:"

    def preload_column_source(self, uuid, path_in_source, source):
        """To save a double-open when using NanoEventsFactory.from_file"""
        key = self.key_root() + tuple_to_key((uuid, path_in_source))
        self._cache[key] = source

    def get_column_handle(self, columnsource, name):
        return columnsource[name]

    def extract_column(self, columnhandle, start, stop):
        # make sure uproot is single-core since our calling context might not be
        return columnhandle.array(
            entry_start=start,
            entry_stop=stop,
            decompression_executor=uproot.source.futures.TrivialExecutor(),
            interpretation_executor=uproot.source.futures.TrivialExecutor(),
        )

    def __len__(self):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError



from coffea.jetmet_tools.JECStack import JECStack
import awkward
import numpy
import warnings
from copy import copy


class CorrectedMETFactory(object):

    def __init__(self, name_map):
        if 'xMETRaw' not in name_map or name_map['xMETRaw'] is None:
            warnings.warn('There is no name mapping for ptRaw,'
                          ' CorrectedJets will assume that <object>.x is raw pt!')
            name_map['xMETRaw'] = name_map['METx'] + '_raw'
        self.treat_pt_as_raw = 'ptRaw' not in name_map

        if 'yMETRaw' not in name_map or name_map['yMETRaw'] is None:
            warnings.warn('There is no name mapping for massRaw,'
                          ' CorrectedJets will assume that <object>.x is raw pt!')
            name_map['yMETRaw'] = name_map['METy'] + '_raw'

        self.name_map = name_map

    def build(self, MET, corrected_jets, lazy_cache):
        if lazy_cache is None:
            raise Exception('CorrectedMETFactory requires a awkward-array cache to function correctly.')
        if (not isinstance(MET, awkward.highlevel.Array) or
            not isinstance(corrected_jets, awkward.highlevel.Array)):
            raise Exception("'MET' and 'corrected_jets' must be an awkward array of some kind!")

        out = copy(MET)

        form = out[self.name_map['METpt']].layout.form
        length = len(out)

        orig_jets = copy(corrected_jets)
        orig_jets[self.name_map['JetPt']] = orig_jets[self.name_map['ptRaw']]
        orig_jets[self.name_map['JetMass']] = orig_jets[self.name_map['massRaw']]

        out['x_orig'] = getattr(out, self.name_map['METx'])
        out['y_orig'] = getattr(out, self.name_map['METy'])

        out[self.name_map['METpt'] + '_orig'] = out[self.name_map['METpt']]
        out[self.name_map['METphi'] + '_orig'] = out[self.name_map['METphi']]

        def corrected_met_cartesian(met, rawJets, corrJets, dim):
            return met[f'{dim}_orig'] - awkward.sum(getattr(rawJets, dim) - getattr(corrJets, dim), axis=-1)

        def corrected_met_cartesian_unc(met, rawJets, corrJets, dimMET, dimJets):
            return getattr(met, dimMET) - awkward.sum(getattr(rawJets, dimJets) - getattr(corrJets, dimJets), axis=-1)

        out['corrected_met_x'] = awkward.virtual(
            corrected_met_cartesian,
            args=(out, orig_jets, corrected_jets, self.name_map['JETx']),
            length=length, form=form, cache=lazy_cache
        )
        out['corrected_met_y'] = awkward.virtual(
            corrected_met_cartesian,
            args=(out, orig_jets, corrected_jets, self.name_map['JETy']),
            length=length, form=form, cache=lazy_cache
        )

        out[self.name_map['METpt']] = awkward.virtual(
            lambda met: numpy.hypot(met['corrected_met_x'], met['corrected_met_y']),
            args=(out, ),
            length=length,
            form=form,
            cache=lazy_cache
        )
        out[self.name_map['METphi']] = awkward.virtual(
            lambda met: numpy.arctan2(met['corrected_met_y'], met['corrected_met_x']),
            args=(out, ),
            length=length,
            form=form,
            cache=lazy_cache
        )

        def make_unclustered_variant(themet, op, deltaX, deltaY):
            variant = copy(themet)
            variant['corrected_met_x'] = awkward.virtual(
                lambda met: op(out['corrected_met_x'], out[f'{deltaX}']),
                args=(out, ),
                length=length,
                form=form,
                cache=lazy_cache
            )
            variant['corrected_met_y'] = awkward.virtual(
                lambda met: op(out['corrected_met_y'], out[f'{deltaY}']),
                args=(out, ),
                length=length,
                form=form,
                cache=lazy_cache
            )
            variant[self.name_map['METpt']] = awkward.virtual(
                lambda met: numpy.hypot(out['corrected_met_x'], out['corrected_met_y']),
                args=(variant, ),
                length=length,
                form=form,
                cache=lazy_cache
            )
            variant[self.name_map['METphi']] = awkward.virtual(
                lambda met: numpy.arctan2(out['corrected_met_y'], out['corrected_met_x']),
                args=(variant, ),
                length=length,
                form=form,
                cache=lazy_cache
            )
            return variant

        unclus_up = make_unclustered_variant(MET, lambda x, y: x + y,
                                             self.name_map['UnClusteredEnergyDeltaX'],
                                             self.name_map['UnClusteredEnergyDeltaY'])
        unclus_down = make_unclustered_variant(MET, lambda x, y: x - y,
                                               self.name_map['UnClusteredEnergyDeltaX'],
                                               self.name_map['UnClusteredEnergyDeltaY'])
        out['MET_UnclusteredEnergy'] = awkward.zip({'up': unclus_up, 'down': unclus_down},
                                                   depth_limit=1,
                                                   with_name='METSystematic')

        def make_variant(name, variation):
            variant = copy(MET)
            variant['corrected_met_x'] = awkward.virtual(
                corrected_met_cartesian_unc,
                args=(out, orig_jets, variation, self.name_map['METx'], self.name_map['JETx']),
                length=length, form=form, cache=lazy_cache
            )
            variant['corrected_met_y'] = awkward.virtual(
                corrected_met_cartesian_unc,
                args=(out, orig_jets, variation, self.name_map['METy'], self.name_map['JETy']),
                length=length, form=form, cache=lazy_cache
            )
            variant[self.name_map['METpt']] = awkward.virtual(
                lambda met: numpy.hypot(met['corrected_met_x'], met['corrected_met_y']),
                args=(variant, ),
                length=length,
                form=form,
                cache=lazy_cache
            )
            variant[self.name_map['METphi']] = awkward.virtual(
                lambda met: numpy.arctan2(met['corrected_met_y'], met['corrected_met_x']),
                args=(variant, ),
                length=length,
                form=form,
                cache=lazy_cache
            )
            return variant

        for unc in filter(lambda x: x.startswith(('JER', 'JES')), awkward.fields(corrected_jets)):
            up = make_variant(unc, corrected_jets[unc].up)
            down = make_variant(unc, corrected_jets[unc].down)
            out[unc] = awkward.zip({'up': up, 'down': down},
                                   depth_limit=1,
                                   with_name='METSystematic')
        return out

    def uncertainties(self):
        return ['MET_UnclusteredEnergy']



from coffea.nanoevents import transforms
from coffea.nanoevents.util import quote, concat
from .base import BaseSchema, listarray_form, zip_forms, nest_jagged_forms


class TreeMakerSchema(BaseSchema):
    """TreeMaker schema builder

    The TreeMaker schema is built from all branches found in the supplied file, based on
    the naming pattern of the branches. From those arrays, TreeMaker collections are formed
    as collections of branches grouped by name, where:

    FIX ME

    All collections are then zipped into one `base.NanoEvents` record and returned.
    """

    def __init__(self, base_form):
        super().__init__(base_form)
        self._form["contents"] = self._build_collections(self._form["contents"])

    def _build_collections(self, branch_forms):
        # Turn any special classes into the appropriate awkward form
        composite_objects = list(set(k.split("/")[0] for k in branch_forms if "/" in k))
        for objname in composite_objects:
            # grab the * from "objname/objname.*"
            components = set(
                k[2 * len(objname) + 2 :]
                for k in branch_forms
                if k.startswith(objname + "/")
            )
            if components == {
                "fCoordinates.fPt",
                "fCoordinates.fEta",
                "fCoordinates.fPhi",
                "fCoordinates.fE",
            }:
                form = zip_forms(
                    {
                        "pt": branch_forms.pop(f"{objname}/{objname}.fCoordinates.fPt"),
                        "eta": branch_forms.pop(
                            f"{objname}/{objname}.fCoordinates.fEta"
                        ),
                        "phi": branch_forms.pop(
                            f"{objname}/{objname}.fCoordinates.fPhi"
                        ),
                        "energy": branch_forms.pop(
                            f"{objname}/{objname}.fCoordinates.fE"
                        ),
                    },
                    objname,
                    "PtEtaPhiELorentzVector",
                )
                print("PtEtaPhiELorentzVector:", objname)
                branch_forms[objname] = form
            elif components == {
                "fCoordinates.fX",
                "fCoordinates.fY",
                "fCoordinates.fZ",
            }:
                form = zip_forms(
                    {
                        "x": branch_forms.pop(f"{objname}/{objname}.fCoordinates.fX"),
                        "y": branch_forms.pop(f"{objname}/{objname}.fCoordinates.fY"),
                        "z": branch_forms.pop(f"{objname}/{objname}.fCoordinates.fZ"),
                    },
                    objname,
                    "Point",
                )
                print("Point:", objname)
                branch_forms[objname] = form
            else:
                raise ValueError(
                    f"Unrecognized class with split branches: {components}"
                )

        collections = [
            "Electrons",
            "GenElectrons",
            "GenJets",
            "GenJetsAK8",
            "GenMuons",
            "GenParticles",
            "GenTaus",
            "GenVertices",
            "Jets",
            "JetsAK8",
            "JetsAK8_subjets",
            "JetsAK15",
            "Muons",
            "PrimaryVertices",
            "TAPElectronTracks",
            "TAPMuonTracks",
            "TAPPionTracks",
            "Tracks",
        ]
        for cname in collections:
            items = sorted(k for k in branch_forms if k.startswith(cname + "_"))
            if len(items) == 0:
                continue
            if cname == "JetsAK8":
                items = [k for k in items if not k.startswith("JetsAK8_subjets")]
                items.append("JetsAK8_subjetsCounts")
            if cname == "JetsAK8_subjets":
                items = [k for k in items if not k.endswith("Counts")]
            if cname not in branch_forms:
                collection = zip_forms(
                    {k[len(cname) + 1]: branch_forms.pop(k) for k in items}, cname
                )
                branch_forms[cname] = collection
            else:
                collection = branch_forms[cname]
                if not collection["class"].startswith("ListOffsetArray"):
                    raise NotImplementedError(
                        f"{cname} isn't a jagged array, not sure what to do"
                    )
                for item in items:
                    itemname = item[len(cname) + 1 :]
                    collection["content"]["contents"][itemname] = branch_forms.pop(
                        item
                    )["content"]

        nest_jagged_forms(
            branch_forms["JetsAK8"],
            branch_forms.pop("JetsAK8_subjets"),
            "subjetsCounts",
            "subjets",
        )

        return branch_forms

    @property
    def behavior(self):
        """Behaviors necessary to implement this schema"""
        from coffea.nanoevents.methods import base, vector

        behavior = {}
        behavior.update(base.behavior)
        behavior.update(vector.behavior)
        return behavior



from coffea.jetmet_tools.JECStack import JECStack
import awkward
import numpy
import warnings
from copy import copy
from functools import partial


_stack_parts = ['jec', 'junc', 'jer', 'jersf']
_MIN_JET_ENERGY = numpy.array(1e-2, dtype=numpy.float32)
_ONE_F32 = numpy.array(1., dtype=numpy.float32)
_ZERO_F32 = numpy.array(0., dtype=numpy.float32)
_JERSF_FORM = {
    "class": "NumpyArray",
    "inner_shape": [
        3
    ],
    "itemsize": 4,
    "format": "f",
    "primitive": "float32"
}


# we're gonna assume that the first record array we encounter is the flattened data
def rewrap_recordarray(layout, depth, data):
    if isinstance(layout, awkward.layout.RecordArray):
        return lambda: data
    return None


def awkward_rewrap(arr, like_what, gfunc):
    behavior = awkward._util.behaviorof(like_what)
    func = partial(gfunc, data=arr.layout)
    layout = awkward.operations.convert.to_layout(like_what)
    newlayout = awkward._util.recursively_apply(layout, func)
    return awkward._util.wrap(newlayout, behavior=behavior)


def rand_gauss(arr, var):
    item = arr[var]

    def getfunction(layout, depth):
        if (isinstance(layout, awkward.layout.NumpyArray)
            or not isinstance(layout, (awkward.layout.Content,
                                       awkward.partition.PartitionedArray)
           )):
            return lambda: awkward.layout.NumpyArray(numpy.random.normal(size=len(layout)).astype(numpy.float32))
        return None

    out = awkward._util.recursively_apply(
        awkward.operations.convert.to_layout(item),
        getfunction)
    assert out is not None
    return awkward._util.wrap(out, awkward._util.behaviorof(item))


def jer_smear(arr, variation, forceStochastic, ptGenJet, ptJet, etaJet):
    pt_gen = arr[ptGenJet] if not forceStochastic else None
    jetPt = arr[ptJet]

    if isinstance(jetPt, awkward.highlevel.Array):
        def getfunction(layout, depth):
            if (isinstance(layout, awkward.layout.NumpyArray)
                or not isinstance(layout, (awkward.layout.Content,
                                           awkward.partition.PartitionedArray)
               )):
                return lambda: awkward.layout.NumpyArray(numpy.zeros_like(size=len(jetPt), dtype=numpy.float32))
            return None
        if forceStochastic:
            pt_gen = awkward._util.recursively_apply(
                awkward.operations.convert.to_layout(jetPt),
                getfunction)
            pt_gen = awkward._util.wrap(pt_gen, awkward._util.behaviorof(jetPt))
    else:
        raise Exception("'arr' must be an awkward array of some kind!")

    jersmear = arr['jet_energy_resolution'] * arr['jet_resolution_rand_gauss']
    jersf = arr['jet_energy_resolution_scale_factor'][:, variation]
    doHybrid = pt_gen > 0

    detSmear = 1 + (jersf - 1) * (jetPt - pt_gen) / jetPt  # because of awkward1.0#367
    stochSmear = 1 + numpy.sqrt(numpy.maximum(jersf**2 - 1, 0)) * jersmear

    min_jet_pt = _MIN_JET_ENERGY / numpy.cosh(arr[etaJet])
    min_jet_pt_corr = min_jet_pt / jetPt

    smearfact = None
    if isinstance(arr, awkward.highlevel.Array):
        smearfact = awkward.where(doHybrid, detSmear, stochSmear)
        smearfact = awkward.where((smearfact * jetPt) < min_jet_pt,
                                  min_jet_pt_corr,
                                  smearfact)

        def getfunction(layout, depth):
            if (isinstance(layout, awkward.layout.NumpyArray)
                or not isinstance(layout, (awkward.layout.Content,
                                           awkward.partition.PartitionedArray)
               )):
                return lambda: awkward.layout.NumpyArray(smearfact)
            return None
        smearfact = awkward._util.recursively_apply(
            awkward.operations.convert.to_layout(jetPt),
            getfunction)
        smearfact = awkward._util.wrap(smearfact, awkward._util.behaviorof(jetPt))
    else:
        raise Exception("'arr' must be an awkward array of some kind!")

    return smearfact


class CorrectedJetsFactory(object):

    def __init__(self, name_map, jec_stack):
        # from PhysicsTools/PatUtils/interface/SmearedJetProducerT.h#L283
        self.forceStochastic = False

        if 'ptRaw' not in name_map or name_map['ptRaw'] is None:
            warnings.warn('There is no name mapping for ptRaw,'
                          ' CorrectedJets will assume that <object>.pt is raw pt!')
            name_map['ptRaw'] = name_map['JetPt'] + '_raw'
        self.treat_pt_as_raw = 'ptRaw' not in name_map

        if 'massRaw' not in name_map or name_map['massRaw'] is None:
            warnings.warn('There is no name mapping for massRaw,'
                          ' CorrectedJets will assume that <object>.mass is raw pt!')
            name_map['ptRaw'] = name_map['JetMass'] + '_raw'

        total_signature = set()
        for part in _stack_parts:
            attr = getattr(jec_stack, part)
            if attr is not None:
                total_signature.update(attr.signature)

        missing = total_signature - set(name_map.keys())
        if len(missing) > 0:
            raise Exception(f'Missing mapping of {missing} in name_map!' +
                            ' Cannot evaluate jet corrections!' +
                            ' Please supply mappings for these variables!')

        if 'ptGenJet' not in name_map:
            warnings.warn('Input JaggedCandidateArray must have "ptGenJet" in order to apply hybrid JER smearing method. Stochastic smearing will be applied.')
            self.forceStochastic = True

        self.real_sig = [v for k, v in name_map.items()]
        self.name_map = name_map
        self.jec_stack = jec_stack

    def uncertainties(self):
        out = ['JER'] if self.jec_stack.jer is not None else []
        if self.jec_stack.junc is not None:
            out.extend(['JES_{0}'.format(unc) for unc in self.jec_stack.junc.levels])
        return out

    def build(self, jets, lazy_cache):
        if lazy_cache is None:
            raise Exception('CorrectedJetsFactory requires a awkward-array cache to function correctly.')

        fields = None
        out = None
        wrap = None
        VirtualType = None
        form = None
        if isinstance(jets, awkward.highlevel.Array):
            fields = awkward.fields(jets)
            if len(fields) == 0:
                raise Exception('Detected awkward: \'jets\' must have attributes specified by keys!')
            out = awkward.flatten(jets)
            wrap = partial(awkward_rewrap, like_what=jets, gfunc=rewrap_recordarray)
            VirtualType = awkward.virtual
            form = out[self.name_map['ptRaw']].layout.form
        else:
            raise Exception('\'jets\' must be an awkward > 1.0.0 array of some kind!')

        if len(fields) == 0:
            raise Exception('Empty record, please pass a jet object with at least {self.real_sig} defined!')

        # take care of nominal JEC (no JER if available)
        out[self.name_map['JetPt'] + '_orig'] = out[self.name_map['JetPt']]
        out[self.name_map['JetMass'] + '_orig'] = out[self.name_map['JetMass']]
        if self.treat_pt_as_raw:
            out[self.name_map['ptRaw']] = out[self.name_map['JetPt']]
            out[self.name_map['massRaw']] = out[self.name_map['JetMass']]

        jec_name_map = {}
        jec_name_map.update(self.name_map)
        jec_name_map['JetPt'] = jec_name_map['ptRaw']
        jec_name_map['JetMass'] = jec_name_map['ptRaw']
        if self.jec_stack.jec is not None:
            jec_args = {k: out[jec_name_map[k]] for k in self.jec_stack.jec.signature}
            out['jet_energy_correction'] = self.jec_stack.jec.getCorrection(**jec_args, form=form, lazy_cache=lazy_cache)
        else:
            out['jet_energy_correction'] = awkward.ones_like(out[self.name_map['JetPt']])

        # finally the lazy binding to the JEC
        def jec_var_corr(arr, varName):
            return arr['jet_energy_correction'] * arr[varName]

        init_pt = partial(VirtualType, jec_var_corr, args=(out, self.name_map['ptRaw']), cache=lazy_cache)
        init_mass = partial(VirtualType, jec_var_corr, args=(out, self.name_map['massRaw']), cache=lazy_cache)

        out[self.name_map['JetPt']] = init_pt(length=len(out), form=form)
        out[self.name_map['JetMass']] = init_mass(length=len(out), form=form)

        out[self.name_map['JetPt'] + '_jec'] = init_pt(length=len(out), form=form)
        out[self.name_map['JetMass'] + '_jec'] = init_mass(length=len(out), form=form)

        # in jer we need to have a stash for the intermediate JEC products
        has_jer = False
        if self.jec_stack.jer is not None and self.jec_stack.jersf is not None:
            has_jer = True
            jer_name_map = {}
            jer_name_map.update(self.name_map)
            jer_name_map['JetPt'] = jer_name_map['JetPt'] + '_jec'
            jer_name_map['JetMass'] = jer_name_map['JetMass'] + '_jec'

            jerargs = {k: out[jer_name_map[k]] for k in self.jec_stack.jer.signature}
            out['jet_energy_resolution'] = self.jec_stack.jer.getResolution(**jerargs, form=form, lazy_cache=lazy_cache)

            jersfargs = {k: out[jer_name_map[k]] for k in self.jec_stack.jersf.signature}
            out['jet_energy_resolution_scale_factor'] = self.jec_stack.jersf.getScaleFactor(**jersfargs, form=_JERSF_FORM, lazy_cache=lazy_cache)

            init_rand_gauss = partial(VirtualType, args=(out, jer_name_map['JetPt']), cache=lazy_cache)
            out['jet_resolution_rand_gauss'] = init_rand_gauss(rand_gauss, length=len(out), form=form)

            init_jerc = partial(VirtualType,
                                jer_smear,
                                args=(out, 0, self.forceStochastic,
                                      jer_name_map['ptGenJet'],
                                      jer_name_map['JetPt'],
                                      jer_name_map['JetEta']),
                                cache=lazy_cache
                                )
            out['jet_energy_resolution_correction'] = init_jerc(length=len(out), form=form)

            def jer_smeared_val(arr, base, varName):
                return arr['jet_energy_resolution_correction'] * base[varName]

            init_pt_jer = partial(VirtualType, jer_smeared_val, args=(out, out, jer_name_map['JetPt']), cache=lazy_cache)
            init_mass_jer = partial(VirtualType, jer_smeared_val, args=(out, out, jer_name_map['JetPt']), cache=lazy_cache)
            out[self.name_map['JetPt']] = init_pt_jer(length=len(out), form=form)
            out[self.name_map['JetMass']] = init_mass_jer(length=len(out), form=form)

            out[self.name_map['JetPt'] + '_jer'] = init_pt_jer(length=len(out), form=form)
            out[self.name_map['JetMass'] + '_jer'] = init_mass_jer(length=len(out), form=form)

            # JER systematics
            jerc_up = partial(VirtualType,
                              jer_smear,
                              args=(out, 1, self.forceStochastic,
                                    jer_name_map['ptGenJet'],
                                    jer_name_map['JetPt'],
                                    jer_name_map['JetEta']),
                              cache=lazy_cache
                              )
            up = awkward.flatten(jets)
            up['jet_energy_resolution_correction'] = jerc_up(length=len(out), form=form)
            init_pt_jer = partial(VirtualType, jer_smeared_val, args=(up, out, jer_name_map['JetPt']), cache=lazy_cache)
            init_mass_jer = partial(VirtualType, jer_smeared_val, args=(up, out, jer_name_map['JetPt']), cache=lazy_cache)
            up[self.name_map['JetPt']] = init_pt_jer(length=len(out), form=form)
            up[self.name_map['JetMass']] = init_mass_jer(length=len(out), form=form)

            jerc_down = partial(VirtualType,
                                jer_smear,
                                args=(out, 2, self.forceStochastic,
                                      jer_name_map['ptGenJet'],
                                      jer_name_map['JetPt'],
                                      jer_name_map['JetEta']),
                                cache=lazy_cache
                                )
            down = awkward.flatten(jets)
            down['jet_energy_resolution_correction'] = jerc_down(length=len(out), form=form)
            init_pt_jer = partial(VirtualType, jer_smeared_val, args=(down, out, jer_name_map['JetPt']), cache=lazy_cache)
            init_mass_jer = partial(VirtualType, jer_smeared_val, args=(down, out, jer_name_map['JetPt']), cache=lazy_cache)
            down[self.name_map['JetPt']] = init_pt_jer(length=len(out), form=form)
            down[self.name_map['JetMass']] = init_mass_jer(length=len(out), form=form)
            out['JER'] = awkward.zip({'up': up, 'down': down}, depth_limit=1, with_name='JetSystematic')

        if self.jec_stack.junc is not None:
            juncnames = {}
            juncnames.update(self.name_map)
            if has_jer:
                juncnames['JetPt'] = juncnames['JetPt'] + '_jer'
                juncnames['JetMass'] = juncnames['JetMass'] + '_jer'
            else:
                juncnames['JetPt'] = juncnames['JetPt'] + '_jec'
                juncnames['JetMass'] = juncnames['JetMass'] + '_jec'
            juncargs = {k: out[juncnames[k]] for k in self.jec_stack.junc.signature}
            juncs = self.jec_stack.junc.getUncertainty(**juncargs)
            for name, func in juncs:
                out[f'jet_energy_uncertainty_{name}'] = func

                def junc_smeared_val(arr, up_down, uncname, varName):
                    return arr[f'jet_energy_uncertainty_{uncname}'][:, up_down] * arr[varName]

                up = awkward.flatten(jets)
                up[self.name_map['JetPt']] = VirtualType(junc_smeared_val,
                                                         args=(out, 0, name,
                                                               juncnames['JetPt'],
                                                               ),
                                                         length=len(out),
                                                         form=form,
                                                         cache=lazy_cache)
                up[self.name_map['JetMass']] = VirtualType(junc_smeared_val,
                                                           args=(out, 0, name,
                                                                 juncnames['JetMass'],
                                                                 ),
                                                           length=len(out),
                                                           form=form,
                                                           cache=lazy_cache)

                down = awkward.flatten(jets)
                down[self.name_map['JetPt']] = VirtualType(junc_smeared_val,
                                                           args=(out, 1, name,
                                                                 juncnames['JetPt'],
                                                                 ),
                                                           length=len(out),
                                                           form=form,
                                                           cache=lazy_cache)
                down[self.name_map['JetMass']] = VirtualType(junc_smeared_val,
                                                             args=(out, 1, name,
                                                                   juncnames['JetMass'],
                                                                   ),
                                                             length=len(out),
                                                             form=form,
                                                             cache=lazy_cache)
                out[f'JES_{name}'] = awkward.zip({'up': up, 'down': down}, depth_limit=1, with_name='JetSystematic')

        return wrap(out)



from coffea import hist, processor
import awkward as ak
import numpy as np
import pandas as pd


class NanoTestProcessorPandas(processor.ProcessorABC):
    def __init__(self, columns=[]):
        self._columns = columns

    @property
    def columns(self):
        return self._columns

    @property
    def accumulator(self):
        return self._accumulator

    def process(self, events):
        output = pd.DataFrame()

        events = events[ak.any(events.Muon.pt > 20, axis=-1)]

        output['run'] = ak.to_numpy(events.run)
        output['event'] = ak.to_numpy(events.event)
        output['dataset'] = events.metadata['dataset']

        output['mu1_pt'] = -999.0
        output['mu2_pt'] = -999.0

        muons = events.Muon[events.Muon.pt > 20]

        counts = ak.num(muons)
        one_muon = ak.to_numpy(counts > 0)
        two_muons = ak.to_numpy(counts > 1)

        output.loc[one_muon, 'mu1_pt'] = ak.to_numpy(muons[one_muon][:, 0].pt)
        output.loc[two_muons, 'mu2_pt'] = ak.to_numpy(muons[two_muons][:, 1].pt)
        return output

    def postprocess(self, accumulator):
        return accumulator



from collections.abc import Mapping


class CachedMapping(Mapping):
    """A cache-wrapped mapping
    Reads will call into ``cache`` first, and if no key exists,
    the read will fall back to ``base``, saving the reult into ``cache``.
    """

    def __init__(self, cache, base):
        self.cache = cache
        self.base = base
        self.stats = {"hit": 0, "miss": 0}

    def __getitem__(self, key):
        try:
            value = self.cache[key]
            self.stats["hit"] += 1
            return value
        except KeyError:
            value = self.base[key]
            self.cache[key] = value
            self.stats["miss"] += 1
            return value

    def __iter__(self):
        return iter(self.base)

    def __len__(self):
        return len(self.base)
